// SNF4CGP/WorkerPool.cpp
// Copyright (C) 2009 ARM Research Labs, LLC.
// See www.armresearch.com for more information.

#include "WorkerPool.hpp"
#include "JobPool.hpp"

#include "../CodeDweller/faults.hpp"

using namespace std;

//// Worker ////////////////////////////////////////////////////////////////////

const ThreadType Worker::Type("Worker");
const ThreadState Worker::Waiting("Waiting");
const ThreadState Worker::Working("Working");

Worker::Worker(WorkerPool& W) :                                                 // Initialize and start the thread.
  Workers(W),
  TimeToStop(false),
  myJob_(0) {
      run();
}

Worker::~Worker() {                                                             // Cleanup on the way down.
    try { stop(); }
    catch(...) {}
}

bool Worker::isJob() {                                                          // Job pointer check.
    return(0 != myJob_);
}

RuntimeCheck CheckForValidJobPointer("Worker::myJob() Check(0 != myJob_)");

Job& Worker::myJob() {                                                          // Safe access to myJob.
    CheckForValidJobPointer(0 != myJob_);
    return(*myJob_);
}

void Worker::clearJob() {                                                       // Forget about the job.
    myJob_ = 0;
}

LogicCheck CheckForOneJobAtATime("Worker::doJob() Check(false == isJob())");

void Worker::doJob(Job& J) {                                                    // Give this worker a job to do.
    ScopeMutex CallTheBall(Busy);
    CheckForOneJobAtATime(false == isJob());
    myJob_ = &J;
    StoppingPoint.produce();
}

LogicCheck CheckForNoJobsAtStop("Worker::stop() Check(false == isJob())");

bool Worker::trySetTimeToStop() {                                               // Statefully set time to stop.
    bool SuccessFlag = false;
    ScopeMutex CallTheBall(Busy);                                               // Don't stop while busy.
    if(false == TimeToStop) {                                                   // Do this only if not done yet.
        CheckForNoJobsAtStop(false == isJob());
        TimeToStop = true;
        SuccessFlag = true;
    }
    return SuccessFlag;
}

void Worker::stop() {                                                           // Stop the worker thread.
    if(trySetTimeToStop()) {                                                    // First try - set time to stop.
        StoppingPoint.produce();                                                // Release the hounds.
        join();                                                                 // Wait for the end.
    }                                                                           // Later tries do nothing.
}

bool Worker::doWork() {                                                         // Get the job done.
    ScopeMutex CallTheBall(Busy);                                               // We're busy - don't touch.
    if(isJob()) {                                                               // Only do a job if we have one.
        myJob().doIt();
        clearJob();
        return true;
    }
    return false;
}

void Worker::myTask() {                                                         // Task for the thread. (my loop).
    for(;;) {
        CurrentThreadState(Waiting);
        StoppingPoint.consume();                                                // Wait for a job.
        CurrentThreadState(Working);
        if(doWork()) {                                                          // After we do a job
            Workers.drop(*this);                                                // drop ourselves back in the pool
            continue;                                                           // and look for another job.
        }                                                                       // If we didn't have a real job
        else break;                                                             // then it's time to stop.
    }
}

//// WorkerPool ////////////////////////////////////////////////////////////////

WorkerPool::WorkerPool() :                                                      // Simple construction.
  AllocatedWorkers(0),
  Started(false) {
}

WorkerPool::~WorkerPool() {                                                     // Cleanup on the way out.
    try { stop(); }
    catch(...) {}
}

void WorkerPool::init() {                                                       // Initialize the worker pool.
    ScopeMutex Busy(AllocationMutex);
    Worker* FirstWorker = makeWorker();                                         // Make the first worker and
    RecycledWorkers.give(FirstWorker);                                          // drop it into the pool.
    Started = true;                                                             // We are now initialized.
}

Worker* WorkerPool::makeWorker() {                                              // Make and count workers.
    Worker* NewWorker = 0;
    NewWorker = new Worker(*this);                                              // Allocate the worker.
    ++AllocatedWorkers;                                                         // If successful count it.
    return NewWorker;                                                           // Return a reference.
}

LogicCheck CheckGrabOnlyWhenStarted("WorkerPool::grab() Check(Started)");
RuntimeCheck CheckForValidGrabbedWorker("WorkerPool::grab() Check(0 != GrabbedWorker)");

Worker& WorkerPool::grab() {                                                    // Grab a worker from the pool.
    ScopeMutex Busy(AllocationMutex);
    CheckGrabOnlyWhenStarted(Started);
    Worker* GrabbedWorker = 0;
    if(0 < RecycledWorkers.size()) GrabbedWorker = RecycledWorkers.take();      // Prefer to use recycled workers.
    else GrabbedWorker = makeWorker();                                          // Make new ones if needed.
    CheckForValidGrabbedWorker(0 != GrabbedWorker);
    return (*GrabbedWorker);
}

// Note that drop() does not use the AllocationMutex. This decouples (somewhat) drop operations from allocation
// operations and improves the performance potential. Since only one thread will ever use the AllocationMutex
// it should never see contention -- it is essentially a safety device in case some later modification gets
// things wrong or some other event occurs that causes a conflict.

void WorkerPool::drop(Worker& W) {                                              // Drop a worker into the pool.
    RecycledWorkers.give(&W);
}

void WorkerPool::killWorkerFromPool() {                                         // Destroy and count workers.
        Worker* WorkerToKill = 0;
        WorkerToKill = RecycledWorkers.take();                                  // Take a worker from the pool.
        WorkerToKill->stop();                                                   // Stop the worker thread.
        delete WorkerToKill;                                                    // Destroy it.
        --AllocatedWorkers;                                                     // Count each one.
}

void WorkerPool::stop() {                                                       // Shutdown the pool (reclaim all)
    ScopeMutex Busy(AllocationMutex);                                           // No grabbing.
    while(0 < AllocatedWorkers) killWorkerFromPool();                           // Kill each worker in the pool.
    Started = false;                                                            // We're not alive now.
}
